from __future__ import absolute_import, division, print_function, unicode_literals

from cgi import escape

from . import models


class SimpleTextBox(models.Content):
    type_name = 'simple-text-box'

    def _render_html(self):
        return escape(self.data['message'], quote=True)


def all_content_types(StartingClass=models.Content):
    """
    :param:
    :return: a dict of str type_name => Content subclass
    """
    class_map = dict()
    for cls in StartingClass.__subclasses__():
        if not isinstance(cls.type_name, models.NoContentTypeNameDefined):
            class_map[cls.type_name] = cls
        class_map.update(all_content_types(cls))
    return class_map





from __future__ import absolute_import, division, print_function, unicode_literals

import os.path
import shutil

from nose.tools import assert_equal, assert_raises

from gazette.content_types import SimpleTextBox
from gazette.models import Content, content_full_name


class TestModels(object):
    def test_abstract(self):
        with assert_raises(NotImplementedError):
            Content('foo').presentation_html()
        with assert_raises(NotImplementedError):
            Content('foo')._render_html()

    def test_item(self):
        ci = SimpleTextBox('foo')


class TestModelStorage(object):
    this_dir = os.path.dirname(os.path.realpath(__file__))
    data_dir = os.path.join(this_dir, 'data')

    def setUp(self):
        # in-memory (isn't very realistic)
        # import datastore.core
        # from .utils import AutoKeyClassDatastore
        # Content.ds = AutoKeyClassDatastore(datastore.DictDatastore())

        import datastore.filesystem
        import datastore.core.serialize
        import jsonpickle

        jsonpickle.set_encoder_options('json', sort_keys=True, indent=2)
        Content.ds = datastore.core.SerializerShimDatastore(
            datastore.filesystem.FileSystemDatastore(self.data_dir),
            serializer=jsonpickle,  # conforms to dumps/loads protocol
        )

    def tearDown(self):
        shutil.rmtree(self.data_dir)


    def test_get_item_missing(self):
        assert_equal(Content.get_by_fullname('/simple-text-box/foobar'),
                     None)

    def test_get_item_missing__show_empty(self):
        assert_equal(Content.get_by_fullname('/simple-text-box/foobar', show_empty=True).presentation_html(),
                     '<div class="gazette" data-path="/simple-text-box/foobar">\nThere is no content created here yet.\n</div>')

    def test_get_item_missing__show_empty_str(self):
        assert_equal(Content.get_by_fullname('/simple-text-box/foobar', show_empty='No content yet').presentation_html(),
                     '<div class="gazette" data-path="/simple-text-box/foobar">\nNo content yet\n</div>')

    def test_set_item_get_item(self):
        ci = SimpleTextBox('foo', data={'bar': 'baz'})
        ci.save()
        assert_equal(SimpleTextBox.get('foo').name,
                     ci.name)
        assert_equal(SimpleTextBox.get('foo').data,
                     ci.data)
        assert_equal(Content.get_by_fullname('/simple-text-box/foo').name,
                     ci.name)
        assert_equal(Content.get_by_fullname('/simple-text-box/foo').data,
                     ci.data)

        assert_equal(Content.get_by_fullname('simple-text-box/foo').name,
                     ci.name)


    def test_query(self):
        # this is just for illustration purposes to show how queries can be done

        SimpleTextBox('foo', data={'bar': 'baz'}).save()
        SimpleTextBox('foo2', data={'bar': 'hey hey hey'}).save()
        SimpleTextBox('foo3', data={'bar': 'baz'}).save()

        from datastore.core.query import Query
        from datastore.core import Key
        import jsonpickle

        def json_getattr(obj, field):
            return Query.object_getattr(jsonpickle.loads(obj), field)

        q = Query(Key('/simple-text-box/'), limit=10, object_getattr=json_getattr)
        q.filter('name', '=', 'foo2')
        result = list(Content.ds.query(q))
        assert_equal(len(result), 1)
        assert_equal(result[0].data['bar'], 'hey hey hey')

        q = Query(Key('/simple-text-box/'), limit=10, object_getattr=json_getattr)
        q.order('-name')
        assert_equal([ci.name for ci in Content.ds.query(q)],
                     ['foo3', 'foo2', 'foo'])

        """
        # TODO: need custom object_getattr for nested access, e.g. into data fields
        q = Query(Key('/simple-text-box/'), limit=10,
                  object_getattr=lambda obj, field: Query.object_getattr(jsonpickle.loads(obj), field))
        q.filter('data.bar', '=', 'baz')
        assert_equal(len(list(Content.ds.query(q))), 2)
        """


class TestContentFullName(object):
    def test_content_full_name__cls_str(self):
        assert_equal(content_full_name(SimpleTextBox, 'abc'),
                     '/simple-text-box/abc')

    def test_content_full_name__item(self):
        assert_equal(content_full_name(SimpleTextBox('abc')),
                     '/simple-text-box/abc')

    def test_content_full_name__invalid(self):
        with assert_raises(TypeError):
            content_full_name(Content, 'foo')


from __future__ import absolute_import, division, print_function, unicode_literals

import json
import os

import bottle
import jsonpickle

from gazette import __version__
from gazette.content_types import all_content_types
from gazette.models import Content


app = bottle.Bottle()

@app.get('/')
def index():
    return {'welcome': 'This is the Gazette API',
            'version': __version__}


@app.get('/content/<full_name:path>')
def get_content(full_name):
    bottle.response.content_type = 'application/json'
    c = Content.get_by_fullname(full_name)
    if c:
        return jsonpickle.dumps(c, unpicklable=False)
    else:
        bottle.abort(404, 'Content not found {}'.format(full_name))


@app.put('/content/<full_name:path>')
def put_content(full_name):
    bottle.response.content_type = 'application/json'

    if not (bottle.request.content_type == 'application/json'
            or bottle.request.content_type.startswith('application/json;')):
        # if the request coming in is not application/json, then bottle won't parse the input into JSON for us
        bottle.abort(415)
    if not bottle.request.json:
        bottle.abort(400)

    c = Content.get_by_fullname(full_name)
    if c:
        c.data = bottle.request.json
        c.save()
    else:
        try:
            type_name, item_name = full_name.split('/', 1)
        except ValueError:
            bottle.abort(404)
        try:
            Type = all_content_types()[type_name]
        except KeyError:
            bottle.abort(404)
        item = Type(item_name, data=bottle.request.json)
        item.save()
        bottle.abort(201)  # created


@app.get('/types')
def get_types():
    bottle.response.content_type = 'application/json'
    return {
        'types': sorted(all_content_types().keys()),
    }


# TODO: schema validation, and validation api


@app.error(404)
#@app.error(500)  # conflicts with debug mode
def error(error):
    bottle.response.content_type = 'application/json'
    return json.dumps({'error': error._status_code, 'message': error.body})


"""
# TODO: need a way to find content types
# could iterate subclasses, but would need to ensure custom subclasses were imported
@app.get('/type/<type_name>')
def content_type(type_name):
    return {}
"""


# FIXME: this works for running bottle from the command line, but it's not good to run stuff at import-time
# automatically.  Need to push it into the app initialization.

if os.environ.get('GZ_DS_DIR'):
    from gazette.utils import datastore_json_files
    Content.ds = datastore_json_files(os.environ['GZ_DS_DIR'])



from __future__ import absolute_import, division, print_function, unicode_literals

from nose.tools import assert_equal
from webtest import TestApp

from gazette.models import Content
from gazette.content_types import SimpleTextBox
from gazette.api import app as api_app


class TestAPIBase(object):
    def setUp(self):
        api_app.catchall = False  # let errors bubble up, instead of a generic 500 page
        self.app = TestApp(api_app)


class TestAPI(TestAPIBase):
    def test_index(self):
        r = self.app.get('/')
        assert_equal(r.content_type, 'application/json')
        assert_equal(r.json['welcome'], 'This is the Gazette API')

    def test_404(self):
        r = self.app.get('/wrong', status=404)
        assert_equal(r.content_type, 'application/json')
        assert_equal(r.json['error'], 404)

    def test_types(self):
        r = self.app.get('/types')
        assert_equal(r.json, {'types': ['simple-text-box']})


class TestAPIWithDatastore(TestAPIBase):
    def setUp(self):
        super(TestAPIWithDatastore, self).setUp()

        import datastore.core
        from gazette.utils import AutoKeyClassDatastore

        Content.ds = AutoKeyClassDatastore(datastore.DictDatastore())

    def test_get_content_404(self):
        r = self.app.get('/content/article/foo', status=404)

    def test_get_content(self):
        content = SimpleTextBox('foo', data={'message': 'abcdef'})
        content.save()
        r = self.app.get('/content/simple-text-box/foo')
        assert_equal(r.json,
                     {'name': 'foo',
                      'data': {'message': 'abcdef'},
                      'version': None,
                      'content_type': 'simple-text-box',
                     })

    def test_create_content(self):
        self.app.put_json('/content/simple-text-box/new',
                          params={'message': 'A new content has been born!'},
                          status=201,
        )
        item = SimpleTextBox.get('new')
        assert_equal(item.name, 'new')
        assert_equal(item.data['message'], 'A new content has been born!')

    def test_update_content(self):
        content = SimpleTextBox('existing', data={'message': 'xyz'})
        content.save()
        self.app.put_json('/content/simple-text-box/existing',
                          params={'message': 'xyz xyz xyz'},
                          status=200,
        )
        item = SimpleTextBox.get('existing')
        assert_equal(item.name, 'existing')
        assert_equal(item.data['message'], 'xyz xyz xyz')
        # TODO assert version

    def test_put_incomplete_path(self):
        self.app.put_json('/content/simple-text-box', params={'x': 'y'}, status=404)
        self.app.put_json('/content/asdfasdf', params={'x': 'y'}, status=404)
        self.app.put_json('/content/asdfasdf/asdf', params={'x': 'y'}, status=404)

    def test_put_wrong_content_type(self):
        self.app.put('/content/simple-text-box/foo', params={'message': 'xyz'}, status=415)
        self.app.put('/content/simple-text-box/foo',
                     headers={'Content-Type': b'application/json; charset=utf-8'},
                     params='{"message": "xyz"}',
                     status=201)

    def test_put_no_contents(self):
        self.app.put_json('/content/simple-text-box/foo', params={}, status=400)


from __future__ import absolute_import, division, print_function, unicode_literals

from nose.tools import assert_equal, assert_raises, assert_in, assert_not_in
from nose import SkipTest

from gazette.models import Content, JinjaPresentationContent
from gazette.content_types import SimpleTextBox, all_content_types


class TestSimpleTextBox(object):
    def test_name(self):
        s = SimpleTextBox('foo')
        assert_equal(s.type_name, 'simple-text-box')

    def test_presentation(self):
        s = SimpleTextBox('foo',
                          data={'message': 'hello, world! <>"'})
        assert_equal(s.presentation_html(),
                     '<div class="gazette" data-path="/simple-text-box/foo">\n'
                     'hello, world! &lt;&gt;&quot;\n'
                     '</div>')

    def test_presentation_nowrap(self):
        s = SimpleTextBox('foo',
                          data={'message': 'hello, world! <>"'})
        assert_equal(s.presentation_html(decorate_html=False),
                     'hello, world! &lt;&gt;&quot;')


def test_all_content_types():
    assert_equal(all_content_types()['simple-text-box'], SimpleTextBox)


def test_all_content_types_only_concrete():
    class Foo(Content):
        pass

    class FooBar(Foo):
        type_name = 'foobar'

        def presentation_html(self):
            return 'baz'

    assert_not_in(Foo, all_content_types().values())
    assert_in(FooBar, all_content_types().values())


class TestJinjaTemplate(object):
    try:
        import jinja2
    except ImportError:
        raise SkipTest

    def test_render_inline_template(self):
        import jinja2

        class JinjaH1inline(JinjaPresentationContent, Content):
            type_name = 'h1-message'

            @property
            def presentation_template(self):
                return jinja2.Template('<h1>{{ message }}</h1>')

        jh1 = JinjaH1inline('abcd', data={'message': 'hello, world!'})
        assert_equal(jh1.presentation_html(decorate_html=False),
                     '<h1>hello, world!</h1>')
        assert_equal(jh1.presentation_html(),
                     '<div class="gazette" data-path="/h1-message/abcd">\n'
                     '<h1>hello, world!</h1>\n'
                     '</div>')


    def test_render_loaded_template(self):
        import jinja2

        class JinjaH1loaded(JinjaPresentationContent, Content):
            type_name = 'h1-message'

            @property
            def presentation_template(self):
                return 'foo/bar/jinja_h1.html'

        j2env = jinja2.Environment(
            loader=jinja2.DictLoader({
                'foo/bar/jinja_h1.html': '<h1>{{ message }}</h1>',
            })
        )

        jh1 = JinjaH1loaded('abcd', data={'message': 'hello, world!'})
        jh1.j2env = j2env
        assert_equal(jh1.presentation_html(decorate_html=False),
                     '<h1>hello, world!</h1>')
        assert_equal(jh1.presentation_html(),
                     '<div class="gazette" data-path="/h1-message/abcd">\n'
                     '<h1>hello, world!</h1>\n'
                     '</div>')


    def test_abstract(self):
        with assert_raises(NotImplementedError):
            JinjaPresentationContent().presentation_template

    def test_show_empty(self):
        import jinja2

        class JinjaH1inline(JinjaPresentationContent, Content):
            type_name = 'h1-message'

            @property
            def presentation_template(self):
                return jinja2.Template('<h1>{{ message }}</h1>')

        assert_equal(JinjaH1inline.get('foobar', show_empty=True).presentation_html(),
                     '<div class="gazette" data-path="/h1-message/foobar">\nThere is no content created here yet.\n</div>')


from __future__ import absolute_import, division, print_function, unicode_literals

from nose.tools import assert_equal

from gazette import utils
from gazette.content_types import SimpleTextBox


class TestContentDictSerializer(object):
    def test_dumps_Content(self):
        ci = SimpleTextBox('hi', data=dict(message='Hello'))
        assert_equal(utils.ContentDictSerializer.dumps(ci),
                     {'name': 'hi',
                      'data': {'message': 'Hello'},
                      'version': None,
                      'content_type': 'simple-text-box',
                     })

    def test_dumps_other(self):
        other = {'foo': 'bar'}
        assert_equal(utils.ContentDictSerializer.dumps(other), other)

    def test_loads_Content(self):
        loaded = utils.ContentDictSerializer.loads(
            {'key': '/simple-text-box/hi',
             'name': 'hi',
             'data': {'message': 'Hello'},
             'version': None,
             'content_type': 'simple-text-box',
            })
        assert_equal(type(loaded), SimpleTextBox)
        assert_equal(loaded.name, 'hi')
        assert_equal(loaded.data, {'message': 'Hello'})
        assert_equal(loaded.version, None)

    def test_loads_other(self):
        other = {'foo': 'bar'}
        assert_equal(utils.ContentDictSerializer.loads(other), other)

    def test_full_cycle(self):
        ci = SimpleTextBox('hi', data=dict(message='Hello'))
        dumped = utils.ContentDictSerializer.dumps(ci)
        loaded = utils.ContentDictSerializer.loads(dumped)
        assert_equal(type(loaded), SimpleTextBox)
        assert_equal(loaded.name, 'hi')
        assert_equal(loaded.data, {'message': 'Hello'})
        assert_equal(loaded.version, None)


class TestMongoNonWrapDatastore(object):

    def test_put(self):
        import datastore.core
        base_ds = datastore.DictDatastore()
        mongo_wrap_ds = utils.MongoNonWrapDatastore(base_ds)

        mongo_wrap_ds.put(datastore.Key('/abc'),
                          {'foo': 'bar'})
        assert_equal(base_ds.get(datastore.Key('/abc')),
                     {'key': '/abc', 'foo':'bar'})


from __future__ import absolute_import, division, print_function, unicode_literals


class NoDatastoreConfigured(object):
    pass


class NoContentTypeNameDefined(object):
    pass


def content_full_name(cls_or_obj, name_if_cls=None):
    """
    Get the full name of a content item.

    :param cls_or_obj: A ``Content`` subclass or instance.
    :param str item_name: The item's name.  Only required if first parameter was a class (not an object instance).
    :return:
    """
    # TODO: versioning
    if name_if_cls:
        name = name_if_cls
    else:
        name = cls_or_obj.name
    return '/'.join(['', cls_or_obj.type_name, name])


class Content(object):
    """
    A definition of content, particularly the templates to show & edit it.

    Subclasses should be used to define specific types (e.g. "blog post")

    Instances of one of a subclasses will hold actual content data (e.g. "Dec 3 blog entry")
    """

    ds = NoDatastoreConfigured

    type_name = NoContentTypeNameDefined()

    def _render_html(self):
        """
        Render the HTML.  Any decoration, extra handling, etc should be done in ``presentation_html()``
        """
        raise NotImplementedError

    def __init__(self, name, version=None, data=None):
        """
        :param str name:
        :param int version:
        :param dict data:
        :return:
        """
        super(Content, self).__init__()
        self.name = name
        self.version = version
        self.data = data
        assert self.type_name
        self.content_type = self.type_name  # store a copy so its explicit in the API, etc

    def save(self):
        self.ds.put(content_full_name(self, self.name), self)

    @classmethod
    def get(cls, name, show_empty=False):
        """
        Gets a content instance.

        :param str name: short name of the content
        :param str show_empty: content not found, returns a placeholder HTML for content instead of returning ``None``.
            ``show_empty may be ``True`` for default placeholder or a string to specify the placeholder text
        :return: a content instance or ``None``
        """
        return cls.get_by_fullname(content_full_name(cls, name), show_empty=show_empty)

    @classmethod
    def get_by_fullname(cls, full_name, show_empty=False):
        """
        Gets a content instance.

        :param str name: full name of the content e.g. ``/article/foo``
        :param str show_empty: content not found, returns a placeholder HTML for content instead of returning ``None``.
            ``show_empty may be ``True`` for default placeholder or a string to specify the placeholder text
        :return: a content instance or ``None``
        """
        if not full_name.startswith('/'):
            full_name = '/' + full_name
        content = cls.ds.get(full_name)

        if not content and show_empty not in (False, None):
            from .content_types import all_content_types
            _, type_name, name = full_name.split('/')
            content_type = all_content_types()[type_name]
            content = content_type(name)
            if show_empty is True:
                placeholder_text = 'There is no content created here yet.'
            else:
                placeholder_text = show_empty
            content._render_html = lambda: placeholder_text

        return content

    def presentation_html(self, decorate_html=True):
        # output the text, no other HTML here
        html = self._render_html()
        if decorate_html:
            html = self._wrap_html(html)
        return html

    def _wrap_html(self, html):
        return self._wrap_html_template(html, content_full_name(self))

    @staticmethod
    def _wrap_html_template(html, full_name):
        return '<div class="gazette" data-path="{}">\n{}\n</div>'.format(full_name, html)


try:
    import jinja2
except ImportError:
    pass
else:

    class JinjaPresentationContent(object):
        '''
        A mixin to implement presentation_html via a template.

        Set the ``j2env`` attribute at either the class or instance level to specify your jinja environment settings
        '''

        j2env = jinja2.Environment()

        def _render_html(self):
            return self.j2env.get_or_select_template(self.presentation_template).render(self.data)

        def presentation_html(self, *args, **kwargs):
            output = super(JinjaPresentationContent, self).presentation_html(*args, **kwargs)
            return jinja2.Markup(output)

        @property
        def presentation_template(self):
            '''
            :return: a template or list of template names to be loaded from the jinja Environment, or a jinja2.Template object
            '''
            raise NotImplementedError



#!/usr/bin/env python

"""Setup script for Gazette."""

import setuptools

from gazette import __project__, __version__

import os
if os.path.exists('README.rst'):
    README = open('README.rst').read()
else:
    README = ""  # a placeholder, readme is generated on release
CHANGES = open('CHANGES.md').read()


setuptools.setup(
    name=__project__,
    version=__version__,

    description="A flexible CMS library that can be added into any app.",
    url='https://sf.net/p/gazette',
    author='Gazette Team',

    packages=setuptools.find_packages(),

    entry_points={'console_scripts': []},

    long_description=(README + '\n' + CHANGES),
    license='MIT',
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
    ],

    install_requires=[
        'bottle',
        'jsonpickle',
    ],

    test_suite='nose.collector',
    tests_require=[
        'nose',
        'jinja2',
        'datastore',
        'webtest',
    ]
)



from __future__ import absolute_import, division, print_function, unicode_literals

import argparse

import jsonpickle
import datastore.core
from datastore.filesystem import FileSystemDatastore
from datastore.core.serialize import SerializerShimDatastore

from gazette.models import Content
from gazette.content_types import all_content_types


def parse_config_args(argv=None):
    """
    Simple arg parser to make common configurations easy.
    If you need a different configuration, you should set it up with a few lines of python to do exactly what you need.

    :param list argv:
    :return: a Datastore
    """
    parser = argparse.ArgumentParser(argv)
    datastores = {
        'fs': FileSystemDatastore,
    }
    serializers = {
        'json': jsonpickle,
    }
    parser.add_argument('--datastore', required=True, choices=datastores.keys())
    parser.add_argument('--path')
    parser.add_argument('--serializer', choices=serializers.keys())
    opts = parser.parse_args(argv)

    ds = datastores[opts.datastore](opts.path)
    if opts.serializer == 'json':
        ds.object_extension = '.json'
    if opts.serializer:
        ds = SerializerShimDatastore(ds, serializer=serializers[opts.serializer])
    return ds


def datastore_json_files(dir):
    """
    Convenience method for a Datastore using local json files

    :param str dir:
    :return: Datastore
    """
    fs_datastore = FileSystemDatastore(dir)
    fs_datastore.object_extension = '.json'
    return SerializerShimDatastore(fs_datastore, serializer=jsonpickle)


class AutoKeyClassDatastore(datastore.core.basic.ShimDatastore):
    '''
    A wrapper around another datastore that automatically converts string "key"
    parameters into instances of the ``Key`` class
    '''

    def delete(self, key):
        if not isinstance(key, datastore.Key):
            key = datastore.Key(key)
        return self.child_datastore.delete(key)

    def get(self, key):
        if not isinstance(key, datastore.Key):
            key = datastore.Key(key)
        return self.child_datastore.get(key)

    def put(self, key, value):
        if not isinstance(key, datastore.Key):
            key = datastore.Key(key)
        return self.child_datastore.put(key, value)

    def query(self, query):
        return self.child_datastore.query(query)


class JsonPickleClean(object):
    @staticmethod
    def dumps(*a, **kw):
        kw.update(unpicklable=False)  # omits "tags" like "py/object": "ClassName"
        return jsonpickle.dumps(*a, **kw)

    @staticmethod
    def loads(*a, **kw):
        return jsonpickle.loads(*a, **kw)


class ContentDictSerializer(object):
    '''
    Serializer/deserializer from Content to dict.

    Perfect for storing into mongo or other datastore that handles dicts.
    '''

    @staticmethod
    def dumps(obj):
        if isinstance(obj, Content):
            return obj.__dict__
        else:
            return obj

    @staticmethod
    def loads(data):
        content_type_name = data.pop('content_type', None)
        if content_type_name:
            content_type = all_content_types()[content_type_name]
            data.pop('key', None)
            return content_type(**data)
        else:
            return data


class MongoNonWrapDatastore(datastore.core.basic.ShimDatastore):
    '''
    Avoids the mongo datastore from wrapping our doc into a nested document,
    by adding the key into the doc itself
    '''

    def get(self, key):
        return self.child_datastore.get(key)

    def put(self, key, value):
        if 'key' not in value:
            value['key'] = str(key)
        return self.child_datastore.put(key, value)

    def delete(self, key):
        return self.child_datastore.delete(key)

    def query(self, query):
        return self.child_datastore.query(query)
